@@ -106,7 +106,8 @@ class Cluster:
             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
         stdout, stderr = process.stdout, process.stderr
 
-        if process.returncode == 4 or not os.listdir(self._data_dir):
+        if (process.returncode == 4 or not os.path.exists(self._data_dir) or
+                not os.listdir(self._data_dir)):
             return 'not-initialized'
         elif process.returncode == 3:
             return 'stopped'
@@ -183,7 +184,7 @@ class Cluster:
             # Make sure server certificate key file has correct permissions.
             keyfile = os.path.join(self._data_dir, 'srvkey.pem')
             shutil.copy(ssl_key, keyfile)
-            os.chmod(keyfile, 0o400)
+            os.chmod(keyfile, 0o600)
             server_settings = server_settings.copy()
             server_settings['ssl_key_file'] = keyfile
 
@@ -202,16 +203,24 @@ class Cluster:
             # of postgres daemon under an Administrative account
             # is not permitted and there is no easy way to drop
             # privileges.
+            if os.getenv('ASYNCPG_DEBUG_SERVER'):
+                stdout = sys.stdout
+            else:
+                stdout = subprocess.DEVNULL
+
             process = subprocess.run(
                 [self._pg_ctl, 'start', '-D', self._data_dir,
                  '-o', ' '.join(extra_args)],
-                stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
-            stderr = process.stderr
+                stdout=stdout, stderr=subprocess.STDOUT)
 
             if process.returncode != 0:
+                if process.stderr:
+                    stderr = ':\n{}'.format(process.stderr.decode())
+                else:
+                    stderr = ''
                 raise ClusterError(
-                    'pg_ctl start exited with status {:d}: {}'.format(
-                        process.returncode, stderr.decode()))
+                    'pg_ctl start exited with status {:d}{}'.format(
+                        process.returncode, stderr))
         else:
             if os.getenv('ASYNCPG_DEBUG_SERVER'):
                 stdout = sys.stdout
@@ -291,6 +300,42 @@ class Cluster:
     def override_connection_spec(self, **kwargs):
         self._connection_spec_override = kwargs
 
+    def reset_wal(self, *, oid=None, xid=None):
+        status = self.get_status()
+        if status == 'not-initialized':
+            raise ClusterError(
+                'cannot modify WAL status: cluster is not initialized')
+
+        if status == 'running':
+            raise ClusterError(
+                'cannot modify WAL status: cluster is running')
+
+        opts = []
+        if oid is not None:
+            opts.extend(['-o', str(oid)])
+        if xid is not None:
+            opts.extend(['-x', str(xid)])
+        if not opts:
+            return
+
+        opts.append(self._data_dir)
+
+        try:
+            reset_wal = self._find_pg_binary('pg_resetwal')
+        except ClusterError:
+            reset_wal = self._find_pg_binary('pg_resetxlog')
+
+        process = subprocess.run(
+            [reset_wal] + opts,
+            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+
+        stderr = process.stderr
+
+        if process.returncode != 0:
+            raise ClusterError(
+                'pg_resetwal exited with status {:d}: {}'.format(
+                    process.returncode, stderr.decode()))
+
     def reset_hba(self):
         """Remove all records from pg_hba.conf."""
         status = self.get_status()
@@ -448,6 +493,7 @@ class Cluster:
                 try:
                     con = loop.run_until_complete(
                         asyncpg.connect(database='postgres',
+                                        user='postgres',
                                         timeout=5, loop=loop,
                                         **self._connection_addr))
                 except (OSError, asyncio.TimeoutError,
