@@ -6,11 +6,14 @@
 
 
 import asyncio
+import asyncpg
 import collections
 import collections.abc
 import itertools
 import struct
+import sys
 import time
+import traceback
 import warnings
 
 from . import compat
@@ -44,7 +47,8 @@ class Connection(metaclass=ConnectionMeta):
                  '_listeners', '_server_version', '_server_caps',
                  '_intro_query', '_reset_query', '_proxy',
                  '_stmt_exclusive_section', '_config', '_params', '_addr',
-                 '_log_listeners', '_cancellations')
+                 '_log_listeners', '_cancellations', '_source_traceback',
+                 '__weakref__')
 
     def __init__(self, protocol, transport, loop,
                  addr: (str, int) or str,
@@ -98,6 +102,27 @@ class Connection(metaclass=ConnectionMeta):
         # `con.execute()`, and `con.executemany()`.
         self._stmt_exclusive_section = _Atomic()
 
+        if loop.get_debug():
+            self._source_traceback = _extract_stack()
+        else:
+            self._source_traceback = None
+
+    def __del__(self):
+        if not self.is_closed() and self._protocol is not None:
+            if self._source_traceback:
+                msg = "unclosed connection {!r}; created at:\n {}".format(
+                    self, self._source_traceback)
+            else:
+                msg = (
+                    "unclosed connection {!r}; run in asyncio debug "
+                    "mode to show the traceback of connection "
+                    "origin".format(self)
+                )
+
+            warnings.warn(msg, ResourceWarning)
+            if not self._loop.is_closed():
+                self.terminate()
+
     async def add_listener(self, channel, callback):
         """Add a listener for Postgres notifications.
 
@@ -112,7 +137,7 @@ class Connection(metaclass=ConnectionMeta):
         """
         self._check_open()
         if channel not in self._listeners:
-            await self.fetch('LISTEN {}'.format(channel))
+            await self.fetch('LISTEN {}'.format(utils._quote_ident(channel)))
             self._listeners[channel] = set()
         self._listeners[channel].add(callback)
 
@@ -127,7 +152,7 @@ class Connection(metaclass=ConnectionMeta):
         self._listeners[channel].remove(callback)
         if not self._listeners[channel]:
             del self._listeners[channel]
-            await self.fetch('UNLISTEN {}'.format(channel))
+            await self.fetch('UNLISTEN {}'.format(utils._quote_ident(channel)))
 
     def add_log_listener(self, callback):
         """Add a listener for Postgres log messages.
@@ -198,12 +223,22 @@ class Connection(metaclass=ConnectionMeta):
         :param deferrable: Specifies whether or not this transaction is
                            deferrable.
 
-        .. _`PostgreSQL documentation`: https://www.postgresql.org/docs/\
-                                        current/static/sql-set-transaction.html
+        .. _`PostgreSQL documentation`:
+                https://www.postgresql.org/docs/
+                current/static/sql-set-transaction.html
         """
         self._check_open()
         return transaction.Transaction(self, isolation, readonly, deferrable)
 
+    def is_in_transaction(self):
+        """Return True if Connection is currently inside a transaction.
+
+        :return bool: True if inside transaction, False otherwise.
+
+        .. versionadded:: 0.16.0
+        """
+        return self._protocol.is_in_transaction()
+
     async def execute(self, query: str, *args, timeout: float=None) -> str:
         """Execute an SQL command (or commands).
 
@@ -286,17 +321,34 @@ class Connection(metaclass=ConnectionMeta):
             stmt_name = ''
 
         statement = await self._protocol.prepare(stmt_name, query, timeout)
-        ready = statement._init_types()
-        if ready is not True:
-            types, intro_stmt = await self.__execute(
-                self._intro_query, (list(ready),), 0, timeout)
-            self._protocol.get_settings().register_data_types(types)
+        need_reprepare = False
+        types_with_missing_codecs = statement._init_types()
+        tries = 0
+        while types_with_missing_codecs:
+            settings = self._protocol.get_settings()
+
+            # Introspect newly seen types and populate the
+            # codec cache.
+            types, intro_stmt = await self._introspect_types(
+                types_with_missing_codecs, timeout)
+
+            settings.register_data_types(types)
+
             # The introspection query has used an anonymous statement,
             # which has blown away the anonymous statement we've prepared
             # for the query, so we need to re-prepare it.
             need_reprepare = not intro_stmt.name and not statement.name
-        else:
-            need_reprepare = False
+            types_with_missing_codecs = statement._init_types()
+            tries += 1
+            if tries > 5:
+                # In the vast majority of cases there will be only
+                # one iteration.  In rare cases, there might be a race
+                # with reload_schema_state(), which would cause a
+                # second try.  More than five is clearly a bug.
+                raise exceptions.InternalClientError(
+                    'could not resolve query result and/or argument types '
+                    'in {} attempts'.format(tries)
+                )
 
         # Now that types have been resolved, populate the codec pipeline
         # for the statement.
@@ -316,6 +368,10 @@ class Connection(metaclass=ConnectionMeta):
 
         return statement
 
+    async def _introspect_types(self, typeoids, timeout):
+        return await self.__execute(
+            self._intro_query, (list(typeoids),), 0, timeout)
+
     def cursor(self, query, *args, prefetch=None, timeout=None):
         """Return a *cursor factory* for the specified query.
 
@@ -441,8 +497,8 @@ class Connection(metaclass=ConnectionMeta):
             >>> asyncio.get_event_loop().run_until_complete(run())
             'COPY 100'
 
-        .. _`COPY statement documentation`: https://www.postgresql.org/docs/\
-                                            current/static/sql-copy.html
+        .. _`COPY statement documentation`:
+            https://www.postgresql.org/docs/current/static/sql-copy.html
 
         .. versionadded:: 0.11.0
         """
@@ -510,8 +566,8 @@ class Connection(metaclass=ConnectionMeta):
             >>> asyncio.get_event_loop().run_until_complete(run())
             'COPY 10'
 
-        .. _`COPY statement documentation`: https://www.postgresql.org/docs/\
-                                            current/static/sql-copy.html
+        .. _`COPY statement documentation`:
+            https://www.postgresql.org/docs/current/static/sql-copy.html
 
         .. versionadded:: 0.11.0
         """
@@ -577,8 +633,8 @@ class Connection(metaclass=ConnectionMeta):
             >>> asyncio.get_event_loop().run_until_complete(run())
             'COPY 140000'
 
-        .. _`COPY statement documentation`: https://www.postgresql.org/docs/\
-                                            current/static/sql-copy.html
+        .. _`COPY statement documentation`:
+            https://www.postgresql.org/docs/current/static/sql-copy.html
 
         .. versionadded:: 0.11.0
         """
@@ -903,12 +959,12 @@ class Connection(metaclass=ConnectionMeta):
         if not typeinfo:
             raise ValueError('unknown type: {}.{}'.format(schema, typename))
 
-        oid = typeinfo['oid']
-        if typeinfo['kind'] != b'b' or typeinfo['elemtype']:
+        if not introspection.is_scalar_type(typeinfo):
             raise ValueError(
                 'cannot use custom codec on non-scalar type {}.{}'.format(
                     schema, typename))
 
+        oid = typeinfo['oid']
         self._protocol.get_settings().add_python_codec(
             oid, typename, schema, 'scalar',
             encoder, decoder, format)
@@ -976,7 +1032,7 @@ class Connection(metaclass=ConnectionMeta):
         :return bool: ``True`` if the connection is closed, ``False``
                       otherwise.
         """
-        return not self._protocol.is_connected() or self._aborted
+        return self._aborted or not self._protocol.is_connected()
 
     async def close(self, *, timeout=None):
         """Close the connection gracefully.
@@ -987,30 +1043,21 @@ class Connection(metaclass=ConnectionMeta):
         .. versionchanged:: 0.14.0
            Added the *timeout* parameter.
         """
-        if self.is_closed():
-            return
-        self._mark_stmts_as_closed()
-        self._listeners.clear()
-        self._log_listeners.clear()
-        self._aborted = True
         try:
-            await self._protocol.close(timeout)
+            if not self.is_closed():
+                await self._protocol.close(timeout)
         except Exception:
             # If we fail to close gracefully, abort the connection.
-            self._aborted = True
-            self._protocol.abort()
+            self._abort()
             raise
         finally:
-            self._clean_tasks()
+            self._cleanup()
 
     def terminate(self):
         """Terminate the connection without waiting for pending data."""
-        self._mark_stmts_as_closed()
-        self._listeners.clear()
-        self._log_listeners.clear()
-        self._aborted = True
-        self._protocol.abort()
-        self._clean_tasks()
+        if not self.is_closed():
+            self._abort()
+        self._cleanup()
 
     async def reset(self, *, timeout=None):
         self._check_open()
@@ -1033,6 +1080,26 @@ class Connection(metaclass=ConnectionMeta):
         if reset_query:
             await self.execute(reset_query, timeout=timeout)
 
+    def _abort(self):
+        # Put the connection into the aborted state.
+        self._aborted = True
+        self._protocol.abort()
+        self._protocol = None
+
+    def _cleanup(self):
+        # Free the resources associated with this connection.
+        # This must be called when a connection is terminated.
+
+        if self._proxy is not None:
+            # Connection is a member of a pool, so let the pool
+            # know that this connection is dead.
+            self._proxy._holder._release_on_close()
+
+        self._mark_stmts_as_closed()
+        self._listeners.clear()
+        self._log_listeners.clear()
+        self._clean_tasks()
+
     def _clean_tasks(self):
         # Wrap-up any remaining tasks associated with this connection.
         if self._cancellations:
@@ -1117,7 +1184,7 @@ class Connection(metaclass=ConnectionMeta):
                 waiter.set_exception(ex)
         finally:
             self._cancellations.discard(
-                asyncio.Task.current_task(self._loop))
+                compat.current_asyncio_task(self._loop))
             if not waiter.done():
                 waiter.set_result(None)
             if w is not None:
@@ -1250,6 +1317,18 @@ class Connection(metaclass=ConnectionMeta):
         else:
             self._drop_local_statement_cache()
 
+    def _drop_local_type_cache(self):
+        self._protocol.get_settings().clear_type_cache()
+
+    def _drop_global_type_cache(self):
+        if self._proxy is not None:
+            # This connection is a member of a pool, so we delegate
+            # the cache drop to the pool.
+            pool = self._proxy._holder._pool
+            pool._drop_type_cache()
+        else:
+            self._drop_local_type_cache()
+
     async def reload_schema_state(self):
         """Indicate that the database schema information must be reloaded.
 
@@ -1292,9 +1371,7 @@ class Connection(metaclass=ConnectionMeta):
 
         .. versionadded:: 0.14.0
         """
-        # It is enough to clear the type cache only once, not in each
-        # connection in the pool.
-        self._protocol.get_settings().clear_type_cache()
+        self._drop_global_type_cache()
         self._drop_global_statement_cache()
 
     async def _execute(self, query, args, limit, timeout, return_status=False):
@@ -1382,7 +1459,7 @@ class Connection(metaclass=ConnectionMeta):
 
 async def connect(dsn=None, *,
                   host=None, port=None,
-                  user=None, password=None,
+                  user=None, password=None, passfile=None,
                   database=None,
                   loop=None,
                   timeout=60,
@@ -1424,6 +1501,11 @@ async def connect(dsn=None, *,
     :param password:
         password used for authentication
 
+    :param passfile:
+        the name of the file used to store passwords
+        (defaults to ``~/.pgpass``, or ``%APPDATA%\postgresql\pgpass.conf``
+        on Windows)
+
     :param loop:
         An asyncio event loop instance.  If ``None``, the default
         event loop will be used.
@@ -1456,7 +1538,8 @@ async def connect(dsn=None, *,
         will be used.
 
     :param dict server_settings:
-        an optional dict of server parameters.
+        an optional dict of server runtime parameters.  Refer to
+        PostgreSQL documentation for a `list of supported options`_.
 
     :param Connection connection_class:
         class of the returned connection object.  Must be a subclass of
@@ -1489,9 +1572,15 @@ async def connect(dsn=None, *,
     .. versionadded:: 0.11.0
        Added ``connection_class`` parameter.
 
+    .. versionadded:: 0.16.0
+       Added ``passfile`` parameter
+       (and support for password files in general).
+
     .. _SSLContext: https://docs.python.org/3/library/ssl.html#ssl.SSLContext
-    .. _create_default_context: https://docs.python.org/3/library/ssl.html#\
-                                ssl.create_default_context
+    .. _create_default_context:
+        https://docs.python.org/3/library/ssl.html#ssl.create_default_context
+    .. _list of supported options:
+        https://www.postgresql.org/docs/current/static/runtime-config.html
     """
     if not issubclass(connection_class, Connection):
         raise TypeError(
@@ -1503,7 +1592,8 @@ async def connect(dsn=None, *,
 
     return await connect_utils._connect(
         loop=loop, timeout=timeout, connection_class=connection_class,
-        dsn=dsn, host=host, port=port, user=user, password=password,
+        dsn=dsn, host=host, port=port, user=user,
+        password=password, passfile=passfile,
         ssl=ssl, database=database,
         server_settings=server_settings,
         command_timeout=command_timeout,
@@ -1726,4 +1816,25 @@ def _detect_server_capabilities(server_version, connection_settings):
     )
 
 
+def _extract_stack(limit=10):
+    """Replacement for traceback.extract_stack() that only does the
+    necessary work for asyncio debug mode.
+    """
+    frame = sys._getframe().f_back
+    try:
+        stack = traceback.StackSummary.extract(
+            traceback.walk_stack(frame), lookup_lines=False)
+    finally:
+        del frame
+
+    apg_path = asyncpg.__path__[0]
+    i = 0
+    while i < len(stack) and stack[i][0].startswith(apg_path):
+        i += 1
+    stack = stack[i:i + limit]
+
+    stack.reverse()
+    return ''.join(traceback.format_list(stack))
+
+
 _uid = 0
