@@ -9,12 +9,15 @@ import datetime
 import decimal
 import ipaddress
 import math
+import os
 import random
 import struct
+import unittest
 import uuid
 
 import asyncpg
 from asyncpg import _testbase as tb
+from asyncpg import cluster as pg_cluster
 
 
 def _timezone(offset):
@@ -22,6 +25,11 @@ def _timezone(offset):
     return datetime.timezone(datetime.timedelta(minutes=minutes))
 
 
+def _system_timezone():
+    d = datetime.datetime.now(datetime.timezone.utc).astimezone()
+    return datetime.timezone(d.utcoffset())
+
+
 infinity_datetime = datetime.datetime(
     datetime.MAXYEAR, 12, 31, 23, 59, 59, 999999)
 negative_infinity_datetime = datetime.datetime(
@@ -29,6 +37,9 @@ negative_infinity_datetime = datetime.datetime(
 
 infinity_date = datetime.date(datetime.MAXYEAR, 12, 31)
 negative_infinity_date = datetime.date(datetime.MINYEAR, 1, 1)
+current_timezone = _system_timezone()
+current_date = datetime.date.today()
+current_datetime = datetime.datetime.now()
 
 
 type_samples = [
@@ -157,6 +168,8 @@ type_samples = [
         negative_infinity_datetime,
         {'textinput': 'infinity', 'output': infinity_datetime},
         {'textinput': '-infinity', 'output': negative_infinity_datetime},
+        {'input': datetime.date(2000, 1, 1),
+         'output': datetime.datetime(2000, 1, 1)}
     ]),
     ('date', 'date', [
         datetime.date(3000, 5, 20),
@@ -182,6 +195,16 @@ type_samples = [
         datetime.datetime(2400, 1, 1, 10, 10, 0, tzinfo=_timezone(2000)),
         infinity_datetime,
         negative_infinity_datetime,
+        {
+            'input': current_date,
+            'output': datetime.datetime(
+                year=current_date.year, month=current_date.month,
+                day=current_date.day, tzinfo=current_timezone),
+        },
+        {
+            'input': current_datetime,
+            'output': current_datetime.replace(tzinfo=current_timezone),
+        }
     ]),
     ('timetz', 'timetz', [
         # timetz retains the offset
@@ -379,6 +402,20 @@ type_samples = [
     ('circle', 'circle', [
         asyncpg.Circle((0.0, 0.0), 100),
     ]),
+    ('tid', 'tid', [
+        (100, 200),
+        (0, 0),
+        (2147483647, 0),
+        (4294967295, 0),
+        (0, 32767),
+        (0, 65535),
+        (4294967295, 65535),
+    ]),
+    ('oid', 'oid', [
+        0,
+        10,
+        4294967295
+    ])
 ]
 
 
@@ -504,6 +541,19 @@ class TestCodecs(tb.ConnectedTestCase):
 
             self.assertEqual(str(res), case)
 
+        try:
+            await self.con.execute(
+                '''
+                    CREATE TABLE tab (v numeric(3, 2));
+                    INSERT INTO tab VALUES (0), (1);
+                ''')
+            res = await self.con.fetchval("SELECT v FROM tab WHERE v = $1", 0)
+            self.assertEqual(str(res), '0.00')
+            res = await self.con.fetchval("SELECT v FROM tab WHERE v = $1", 1)
+            self.assertEqual(str(res), '1.00')
+        finally:
+            await self.con.execute('DROP TABLE tab')
+
         res = await self.con.fetchval(
             "SELECT $1::numeric", decimal.Decimal('NaN'))
         self.assertTrue(res.is_nan())
@@ -512,17 +562,19 @@ class TestCodecs(tb.ConnectedTestCase):
             "SELECT $1::numeric", decimal.Decimal('sNaN'))
         self.assertTrue(res.is_nan())
 
-        with self.assertRaisesRegex(ValueError, 'numeric type does not '
-                                                'support infinite values'):
+        with self.assertRaisesRegex(asyncpg.DataError,
+                                    'numeric type does not '
+                                    'support infinite values'):
             await self.con.fetchval(
                 "SELECT $1::numeric", decimal.Decimal('-Inf'))
 
-        with self.assertRaisesRegex(ValueError, 'numeric type does not '
-                                                'support infinite values'):
+        with self.assertRaisesRegex(asyncpg.DataError,
+                                    'numeric type does not '
+                                    'support infinite values'):
             await self.con.fetchval(
                 "SELECT $1::numeric", decimal.Decimal('+Inf'))
 
-        with self.assertRaises(decimal.InvalidOperation):
+        with self.assertRaisesRegex(asyncpg.DataError, 'invalid'):
             await self.con.fetchval(
                 "SELECT $1::numeric", 'invalid')
 
@@ -548,18 +600,18 @@ class TestCodecs(tb.ConnectedTestCase):
 
     async def test_invalid_input(self):
         cases = [
-            ('bytea', TypeError, 'a bytes-like object is required', [
+            ('bytea', 'a bytes-like object is required', [
                 1,
                 'aaa'
             ]),
-            ('bool', TypeError, 'a boolean is required', [
+            ('bool', 'a boolean is required', [
                 1,
             ]),
-            ('int2', TypeError, 'an integer is required', [
+            ('int2', 'an integer is required', [
                 '2',
                 'aa',
             ]),
-            ('smallint', OverflowError, 'int too big to be encoded as INT2', [
+            ('smallint', 'value out of int16 range', [
                 2**256,  # check for the same exception for any big numbers
                 decimal.Decimal("2000000000000000000000000000000"),
                 0xffff,
@@ -567,46 +619,82 @@ class TestCodecs(tb.ConnectedTestCase):
                 32768,
                 -32769
             ]),
-            ('float4', ValueError, 'float value too large', [
+            ('float4', 'value out of float32 range', [
                 4.1 * 10 ** 40,
                 -4.1 * 10 ** 40,
             ]),
-            ('int4', TypeError, 'an integer is required', [
+            ('int4', 'an integer is required', [
                 '2',
                 'aa',
             ]),
-            ('int', OverflowError, 'int too big to be encoded as INT4', [
+            ('int', 'value out of int32 range', [
                 2**256,  # check for the same exception for any big numbers
                 decimal.Decimal("2000000000000000000000000000000"),
                 0xffffffff,
                 2**31,
                 -2**31 - 1,
             ]),
-            ('int8', TypeError, 'an integer is required', [
+            ('int8', 'an integer is required', [
                 '2',
                 'aa',
             ]),
-            ('bigint', OverflowError, 'int too big to be encoded as INT8', [
+            ('bigint', 'value out of int64 range', [
                 2**256,  # check for the same exception for any big numbers
                 decimal.Decimal("2000000000000000000000000000000"),
                 0xffffffffffffffff,
                 2**63,
                 -2**63 - 1,
             ]),
-            ('text', TypeError, 'expected str, got bytes', [
+            ('text', 'expected str, got bytes', [
                 b'foo'
             ]),
-            ('text', TypeError, 'expected str, got list', [
+            ('text', 'expected str, got list', [
                 [1]
-            ])
+            ]),
+            ('tid', 'list or tuple expected', [
+                b'foo'
+            ]),
+            ('tid', 'invalid number of elements in tid tuple', [
+                [],
+                (),
+                [1, 2, 3],
+                (4,),
+            ]),
+            ('tid', 'tuple id block value out of uint32 range', [
+                (-1, 0),
+                (2**256, 0),
+                (0xffffffff + 1, 0),
+                (2**32, 0),
+            ]),
+            ('tid', 'tuple id offset value out of uint16 range', [
+                (0, -1),
+                (0, 2**256),
+                (0, 0xffff + 1),
+                (0, 0xffffffff),
+                (0, 65536),
+            ]),
+            ('oid', 'value out of uint32 range', [
+                2 ** 32,
+                -1,
+            ]),
+            ('timestamp', r"expected a datetime\.date.*got 'str'", [
+                'foo'
+            ]),
+            ('timestamptz', r"expected a datetime\.date.*got 'str'", [
+                'foo'
+            ]),
         ]
 
-        for typname, errcls, errmsg, data in cases:
+        for typname, errmsg, data in cases:
             stmt = await self.con.prepare("SELECT $1::" + typname)
 
             for sample in data:
                 with self.subTest(sample=sample, typname=typname):
-                    with self.assertRaisesRegex(errcls, errmsg):
+                    full_errmsg = (
+                        r'invalid input for query argument \$1:.*' + errmsg)
+
+                    with self.assertRaisesRegex(
+                            asyncpg.DataError, full_errmsg):
                         await stmt.fetchval(sample)
 
     async def test_arrays(self):
@@ -661,33 +749,56 @@ class TestCodecs(tb.ConnectedTestCase):
 
                 self.assertEqual(result, case, err_msg)
 
-        with self.assertRaisesRegex(ValueError, 'dimensions'):
+        # A sized iterable is fine as array input.
+        class Iterable:
+            def __iter__(self):
+                return iter([1, 2, 3])
+
+            def __len__(self):
+                return 3
+
+        result = await self.con.fetchval("SELECT $1::int[]", Iterable())
+        self.assertEqual(result, [1, 2, 3])
+
+        # A pure container is _not_ OK for array input.
+        class SomeContainer:
+            def __contains__(self, item):
+                return False
+
+        with self.assertRaisesRegex(asyncpg.DataError,
+                                    'sized iterable container expected'):
+            result = await self.con.fetchval("SELECT $1::int[]",
+                                             SomeContainer())
+
+        with self.assertRaisesRegex(asyncpg.DataError, 'dimensions'):
             await self.con.fetchval(
                 "SELECT $1::int[]",
                 [[[[[[[1]]]]]]])
 
-        with self.assertRaisesRegex(ValueError, 'non-homogeneous'):
+        with self.assertRaisesRegex(asyncpg.DataError, 'non-homogeneous'):
             await self.con.fetchval(
                 "SELECT $1::int[]",
                 [1, [1]])
 
-        with self.assertRaisesRegex(ValueError, 'non-homogeneous'):
+        with self.assertRaisesRegex(asyncpg.DataError, 'non-homogeneous'):
             await self.con.fetchval(
                 "SELECT $1::int[]",
                 [[1], 1, [2]])
 
-        with self.assertRaisesRegex(ValueError, 'invalid array element'):
+        with self.assertRaisesRegex(asyncpg.DataError,
+                                    'invalid array element'):
             await self.con.fetchval(
                 "SELECT $1::int[]",
                 [1, 't', 2])
 
-        with self.assertRaisesRegex(ValueError, 'invalid array element'):
+        with self.assertRaisesRegex(asyncpg.DataError,
+                                    'invalid array element'):
             await self.con.fetchval(
                 "SELECT $1::int[]",
                 [[1], ['t'], [2]])
 
-        with self.assertRaisesRegex(TypeError,
-                                    'non-trivial iterable expected'):
+        with self.assertRaisesRegex(asyncpg.DataError,
+                                    'sized iterable container expected'):
             await self.con.fetchval(
                 "SELECT $1::int[]",
                 1)
@@ -810,11 +921,11 @@ class TestCodecs(tb.ConnectedTestCase):
                     self.assertEqual(result, expected)
 
         with self.assertRaisesRegex(
-                TypeError, 'list, tuple or Range object expected'):
+                asyncpg.DataError, 'list, tuple or Range object expected'):
             await self.con.fetch("SELECT $1::int4range", 'aa')
 
         with self.assertRaisesRegex(
-                ValueError, 'expected 0, 1 or 2 elements'):
+                asyncpg.DataError, 'expected 0, 1 or 2 elements'):
             await self.con.fetch("SELECT $1::int4range", (0, 2, 3))
 
         cases = [(asyncpg.Range(0, 1), asyncpg.Range(0, 1), 1),
@@ -856,7 +967,8 @@ class TestCodecs(tb.ConnectedTestCase):
 
             self.assertEqual(res, {'foo': '2', 'bar': '3'})
 
-            with self.assertRaisesRegex(ValueError, 'null value not allowed'):
+            with self.assertRaisesRegex(asyncpg.DataError,
+                                        'null value not allowed'):
                 await self.con.fetchval('''
                     SELECT $1::hstore AS result
                 ''', {None: '1'})
@@ -1017,6 +1129,40 @@ class TestCodecs(tb.ConnectedTestCase):
                 DROP EXTENSION hstore
             ''')
 
+    async def test_custom_codec_on_domain(self):
+        """Test encoding/decoding using a custom codec on a domain."""
+        await self.con.execute('''
+            CREATE DOMAIN custom_codec_t AS int
+        ''')
+
+        try:
+            await self.con.set_type_codec(
+                'custom_codec_t',
+                encoder=lambda v: str(v),
+                decoder=lambda v: int(v))
+
+            v = await self.con.fetchval('SELECT $1::custom_codec_t', 10)
+            self.assertEqual(v, 10)
+        finally:
+            await self.con.execute('DROP DOMAIN custom_codec_t')
+
+    async def test_custom_codec_on_enum(self):
+        """Test encoding/decoding using a custom codec on an enum."""
+        await self.con.execute('''
+            CREATE TYPE custom_codec_t AS ENUM ('foo', 'bar', 'baz')
+        ''')
+
+        try:
+            await self.con.set_type_codec(
+                'custom_codec_t',
+                encoder=lambda v: str(v).lstrip('enum :'),
+                decoder=lambda v: 'enum: ' + str(v))
+
+            v = await self.con.fetchval('SELECT $1::custom_codec_t', 'foo')
+            self.assertEqual(v, 'enum: foo')
+        finally:
+            await self.con.execute('DROP TYPE custom_codec_t')
+
     async def test_custom_codec_override_binary(self):
         """Test overriding core codecs."""
         import json
@@ -1446,3 +1592,36 @@ class TestCodecs(tb.ConnectedTestCase):
     async def test_no_result(self):
         st = await self.con.prepare('rollback')
         self.assertTupleEqual(st.get_attributes(), ())
+
+
+@unittest.skipIf(os.environ.get('PGHOST'), 'using remote cluster for testing')
+class TestCodecsLargeOIDs(tb.ConnectedTestCase):
+    @classmethod
+    def setup_cluster(cls):
+        cls.cluster = cls.new_cluster(pg_cluster.TempCluster)
+        cls.cluster.reset_wal(oid=2147483648)
+        cls.start_cluster(cls.cluster)
+
+    async def test_custom_codec_large_oid(self):
+        await self.con.execute('CREATE DOMAIN test_domain_t AS int')
+        try:
+            oid = await self.con.fetchval('''
+                SELECT oid FROM pg_type WHERE typname = 'test_domain_t'
+            ''')
+            self.assertEqual(oid, 2147483648)
+
+            # Test that introspection handles large OIDs
+            v = await self.con.fetchval('SELECT $1::test_domain_t', 10)
+            self.assertEqual(v, 10)
+
+            # Test that custom codec logic handles large OIDs
+            await self.con.set_type_codec(
+                'test_domain_t',
+                encoder=lambda v: str(v),
+                decoder=lambda v: int(v))
+
+            v = await self.con.fetchval('SELECT $1::test_domain_t', 10)
+            self.assertEqual(v, 10)
+
+        finally:
+            await self.con.execute('DROP DOMAIN test_domain_t')
