@@ -7,11 +7,16 @@
 
 import asyncio
 import contextlib
+import gc
 import ipaddress
 import os
 import platform
 import ssl
+import stat
+import tempfile
+import textwrap
 import unittest
+import weakref
 
 import asyncpg
 from asyncpg import _testbase as tb
@@ -354,6 +359,7 @@ class TestConnectParams(tb.TestCase):
         port = testcase.get('port')
         host = testcase.get('host')
         password = testcase.get('password')
+        passfile = testcase.get('passfile')
         database = testcase.get('database')
         server_settings = testcase.get('server_settings')
 
@@ -377,8 +383,8 @@ class TestConnectParams(tb.TestCase):
 
             addrs, params = connect_utils._parse_connect_dsn_and_args(
                 dsn=dsn, host=host, port=port, user=user, password=password,
-                database=database, ssl=None, connect_timeout=None,
-                server_settings=server_settings)
+                passfile=passfile, database=database, ssl=None,
+                connect_timeout=None, server_settings=server_settings)
 
             params = {k: v for k, v in params._asdict().items()
                       if v is not None}
@@ -431,6 +437,230 @@ class TestConnectParams(tb.TestCase):
         for testcase in self.TESTS:
             self.run_testcase(testcase)
 
+    def test_connect_pgpass_regular(self):
+        passfile = tempfile.NamedTemporaryFile('w+t', delete=False)
+        passfile.write(textwrap.dedent(R'''
+            abc:*:*:user:password from pgpass for user@abc
+            localhost:*:*:*:password from pgpass for localhost
+            cde:5433:*:*:password from pgpass for cde:5433
+
+            *:*:*:testuser:password from pgpass for testuser
+            *:*:testdb:*:password from pgpass for testdb
+            # comment
+            *:*:test\:db:test\\:password from pgpass with escapes
+        '''))
+        passfile.close()
+        os.chmod(passfile.name, stat.S_IWUSR | stat.S_IRUSR)
+
+        try:
+            # passfile path in env
+            self.run_testcase({
+                'env': {
+                    'PGPASSFILE': passfile.name
+                },
+                'host': 'abc',
+                'user': 'user',
+                'database': 'db',
+                'result': (
+                    [('abc', 5432)],
+                    {
+                        'password': 'password from pgpass for user@abc',
+                        'user': 'user',
+                        'database': 'db',
+                    }
+                )
+            })
+
+            # passfile path as explicit arg
+            self.run_testcase({
+                'host': 'abc',
+                'user': 'user',
+                'database': 'db',
+                'passfile': passfile.name,
+                'result': (
+                    [('abc', 5432)],
+                    {
+                        'password': 'password from pgpass for user@abc',
+                        'user': 'user',
+                        'database': 'db',
+                    }
+                )
+            })
+
+            # passfile path in dsn
+            self.run_testcase({
+                'dsn': 'postgres://user@abc/db?passfile={}'.format(
+                    passfile.name),
+                'result': (
+                    [('abc', 5432)],
+                    {
+                        'password': 'password from pgpass for user@abc',
+                        'user': 'user',
+                        'database': 'db',
+                    }
+                )
+            })
+
+            self.run_testcase({
+                'host': 'localhost',
+                'user': 'user',
+                'database': 'db',
+                'passfile': passfile.name,
+                'result': (
+                    [('localhost', 5432)],
+                    {
+                        'password': 'password from pgpass for localhost',
+                        'user': 'user',
+                        'database': 'db',
+                    }
+                )
+            })
+
+            if _system != 'Windows':
+                # unix socket gets normalized as localhost
+                self.run_testcase({
+                    'host': '/tmp',
+                    'user': 'user',
+                    'database': 'db',
+                    'passfile': passfile.name,
+                    'result': (
+                        ['/tmp/.s.PGSQL.5432'],
+                        {
+                            'password': 'password from pgpass for localhost',
+                            'user': 'user',
+                            'database': 'db',
+                        }
+                    )
+                })
+
+            # port matching (also tests that `:` can be part of password)
+            self.run_testcase({
+                'host': 'cde',
+                'port': 5433,
+                'user': 'user',
+                'database': 'db',
+                'passfile': passfile.name,
+                'result': (
+                    [('cde', 5433)],
+                    {
+                        'password': 'password from pgpass for cde:5433',
+                        'user': 'user',
+                        'database': 'db',
+                    }
+                )
+            })
+
+            # user matching
+            self.run_testcase({
+                'host': 'def',
+                'user': 'testuser',
+                'database': 'db',
+                'passfile': passfile.name,
+                'result': (
+                    [('def', 5432)],
+                    {
+                        'password': 'password from pgpass for testuser',
+                        'user': 'testuser',
+                        'database': 'db',
+                    }
+                )
+            })
+
+            # database matching
+            self.run_testcase({
+                'host': 'efg',
+                'user': 'user',
+                'database': 'testdb',
+                'passfile': passfile.name,
+                'result': (
+                    [('efg', 5432)],
+                    {
+                        'password': 'password from pgpass for testdb',
+                        'user': 'user',
+                        'database': 'testdb',
+                    }
+                )
+            })
+
+            # test escaping
+            self.run_testcase({
+                'host': 'fgh',
+                'user': R'test\\',
+                'database': R'test\:db',
+                'passfile': passfile.name,
+                'result': (
+                    [('fgh', 5432)],
+                    {
+                        'password': 'password from pgpass with escapes',
+                        'user': R'test\\',
+                        'database': R'test\:db',
+                    }
+                )
+            })
+
+        finally:
+            os.unlink(passfile.name)
+
+    @unittest.skipIf(_system == 'Windows', 'no mode checking on Windows')
+    def test_connect_pgpass_badness_mode(self):
+        # Verify that .pgpass permissions are checked
+        with tempfile.NamedTemporaryFile('w+t') as passfile:
+            os.chmod(passfile.name,
+                     stat.S_IWUSR | stat.S_IRUSR | stat.S_IWGRP | stat.S_IRGRP)
+
+            with self.assertWarnsRegex(
+                    UserWarning,
+                    'password file .* has group or world access'):
+                self.run_testcase({
+                    'host': 'abc',
+                    'user': 'user',
+                    'database': 'db',
+                    'passfile': passfile.name,
+                    'result': (
+                        [('abc', 5432)],
+                        {
+                            'user': 'user',
+                            'database': 'db',
+                        }
+                    )
+                })
+
+    def test_connect_pgpass_badness_non_file(self):
+        # Verify warnings when .pgpass is not a file
+        with tempfile.TemporaryDirectory() as passfile:
+            with self.assertWarnsRegex(
+                    UserWarning,
+                    'password file .* is not a plain file'):
+                self.run_testcase({
+                    'host': 'abc',
+                    'user': 'user',
+                    'database': 'db',
+                    'passfile': passfile,
+                    'result': (
+                        [('abc', 5432)],
+                        {
+                            'user': 'user',
+                            'database': 'db',
+                        }
+                    )
+                })
+
+    def test_connect_pgpass_nonexistent(self):
+        # nonexistent passfile is OK
+        self.run_testcase({
+            'host': 'abc',
+            'user': 'user',
+            'database': 'db',
+            'passfile': 'totally nonexistent',
+            'result': (
+                [('abc', 5432)],
+                {
+                    'user': 'user',
+                    'database': 'db',
+                }
+            )
+        })
+
     async def test_connect_args_validation(self):
         for val in {-1, 'a', True, False, 0}:
             with self.assertRaisesRegex(ValueError, 'greater than 0'):
@@ -528,8 +758,9 @@ class TestSSLConnection(tb.ConnectedTestCase):
 
     @classmethod
     def setup_cluster(cls):
-        cls.cluster = cls.start_cluster(
-            pg_cluster.TempCluster, server_settings=cls.get_server_settings())
+        cls.cluster = cls.new_cluster(pg_cluster.TempCluster)
+        cls.start_cluster(
+            cls.cluster, server_settings=cls.get_server_settings())
 
     def setUp(self):
         super().setUp()
@@ -586,11 +817,17 @@ class TestSSLConnection(tb.ConnectedTestCase):
             await con.close()
 
     async def test_ssl_connection_default_context(self):
-        with self.assertRaisesRegex(ssl.SSLError, 'verify failed'):
-            await self.connect(
-                host='localhost',
-                user='ssl_user',
-                ssl=True)
+        # XXX: uvloop artifact
+        old_handler = self.loop.get_exception_handler()
+        try:
+            self.loop.set_exception_handler(lambda *args: None)
+            with self.assertRaisesRegex(ssl.SSLError, 'verify failed'):
+                await self.connect(
+                    host='localhost',
+                    user='ssl_user',
+                    ssl=True)
+        finally:
+            self.loop.set_exception_handler(old_handler)
 
     async def test_ssl_connection_pool(self):
         ssl_context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
@@ -616,3 +853,44 @@ class TestSSLConnection(tb.ConnectedTestCase):
         tasks = [worker() for _ in range(100)]
         await asyncio.gather(*tasks, loop=self.loop)
         await pool.close()
+
+
+class TestConnectionGC(tb.ClusterTestCase):
+
+    async def _run_no_explicit_close_test(self):
+        con = await self.connect()
+        proto = con._protocol
+        conref = weakref.ref(con)
+        del con
+
+        gc.collect()
+        gc.collect()
+        gc.collect()
+
+        self.assertIsNone(conref())
+        self.assertTrue(proto.is_closed())
+
+    async def test_no_explicit_close_no_debug(self):
+        olddebug = self.loop.get_debug()
+        self.loop.set_debug(False)
+        try:
+            with self.assertWarnsRegex(
+                    ResourceWarning,
+                    r'unclosed connection.*run in asyncio debug'):
+                await self._run_no_explicit_close_test()
+        finally:
+            self.loop.set_debug(olddebug)
+
+    async def test_no_explicit_close_with_debug(self):
+        olddebug = self.loop.get_debug()
+        self.loop.set_debug(True)
+        try:
+            with self.assertWarnsRegex(ResourceWarning,
+                                       r'unclosed connection') as rw:
+                await self._run_no_explicit_close_test()
+
+            msg = rw.warning.args[0]
+            self.assertIn(' created at:\n', msg)
+            self.assertIn('in test_no_explicit_close_with_debug', msg)
+        finally:
+            self.loop.set_debug(olddebug)
