@@ -6,7 +6,6 @@
 
 
 import asyncio
-import asyncpg
 import inspect
 import os
 import platform
@@ -16,6 +15,7 @@ import textwrap
 import time
 import unittest
 
+import asyncpg
 from asyncpg import _testbase as tb
 from asyncpg import connection as pg_connection
 from asyncpg import cluster as pg_cluster
@@ -92,16 +92,23 @@ class TestPool(tb.ConnectedTestCase):
                                       min_size=1, max_size=1)
 
         con = await pool.acquire(timeout=POOL_NOMINAL_TIMEOUT)
-        con.terminate()
-        await pool.release(con)
 
-        async with pool.acquire(timeout=POOL_NOMINAL_TIMEOUT) as con:
-            con.terminate()
+        # Manual termination of pool connections releases the
+        # pool item immediately.
+        con.terminate()
+        self.assertIsNone(pool._holders[0]._con)
+        self.assertIsNone(pool._holders[0]._in_use)
 
         con = await pool.acquire(timeout=POOL_NOMINAL_TIMEOUT)
         self.assertEqual(await con.fetchval('SELECT 1'), 1)
 
-        await pool.close()
+        await con.close()
+        self.assertIsNone(pool._holders[0]._con)
+        self.assertIsNone(pool._holders[0]._in_use)
+        # Calling release should not hurt.
+        await pool.release(con)
+
+        pool.terminate()
 
     async def test_pool_05(self):
         for n in {1, 3, 5, 10, 20, 100}:
@@ -126,7 +133,8 @@ class TestPool(tb.ConnectedTestCase):
         async with self.create_pool(database='postgres',
                                     min_size=5, max_size=5,
                                     setup=setup) as pool:
-            con = await pool.acquire()
+            async with pool.acquire() as con:
+                pass
 
         self.assertIs(con, await fut)
 
@@ -172,9 +180,13 @@ class TestPool(tb.ConnectedTestCase):
         pool2 = await self.create_pool(database='postgres',
                                        min_size=1, max_size=1)
 
-        con = await pool1.acquire(timeout=POOL_NOMINAL_TIMEOUT)
-        with self.assertRaisesRegex(asyncpg.InterfaceError, 'is not a member'):
-            await pool2.release(con)
+        try:
+            con = await pool1.acquire(timeout=POOL_NOMINAL_TIMEOUT)
+            with self.assertRaisesRegex(asyncpg.InterfaceError,
+                                        'is not a member'):
+                await pool2.release(con)
+        finally:
+            await pool1.release(con)
 
         await pool1.close()
         await pool2.close()
@@ -280,7 +292,8 @@ class TestPool(tb.ConnectedTestCase):
             pass
 
         async def setup(con):
-            nonlocal setup_calls
+            nonlocal setup_calls, last_con
+            last_con = con
             setup_calls += 1
             if setup_calls > 1:
                 cons.append(con)
@@ -290,28 +303,32 @@ class TestPool(tb.ConnectedTestCase):
 
         with self.subTest(method='setup'):
             setup_calls = 0
+            last_con = None
             cons = []
             async with self.create_pool(database='postgres',
                                         min_size=1, max_size=1,
                                         setup=setup) as pool:
                 with self.assertRaises(Error):
                     await pool.acquire()
+                self.assertTrue(last_con.is_closed())
 
-                con = await pool.acquire()
-                self.assertEqual(cons, ['error', con])
+                async with pool.acquire() as con:
+                    self.assertEqual(cons, ['error', con])
 
         with self.subTest(method='init'):
             setup_calls = 0
+            last_con = None
             cons = []
             async with self.create_pool(database='postgres',
                                         min_size=0, max_size=1,
                                         init=setup) as pool:
                 with self.assertRaises(Error):
                     await pool.acquire()
+                self.assertTrue(last_con.is_closed())
 
-                con = await pool.acquire()
-                self.assertEqual(await con.fetchval('select 1::int'), 1)
-                self.assertEqual(cons, ['error', con._con])
+                async with pool.acquire() as con:
+                    self.assertEqual(await con.fetchval('select 1::int'), 1)
+                    self.assertEqual(cons, ['error', con._con])
 
     async def test_pool_auth(self):
         if not self.cluster.is_managed():
@@ -665,11 +682,10 @@ class TestPool(tb.ConnectedTestCase):
         true_con.terminate()
 
         # now pool should reopen terminated connection
-        con = await pool.acquire(timeout=POOL_NOMINAL_TIMEOUT)
+        async with pool.acquire(timeout=POOL_NOMINAL_TIMEOUT) as con:
+            self.assertEqual(await con.fetchval('SELECT 1'), 1)
+            await con.close()
 
-        self.assertEqual(await con.fetchval('SELECT 1'), 1)
-
-        await con.close()
         await pool.close()
 
     @unittest.skipIf(sys.version_info[:2] < (3, 6), 'no asyncgen support')
@@ -741,19 +757,143 @@ class TestPool(tb.ConnectedTestCase):
                     async for _ in iterate(con):  # noqa
                         raise MyException()
 
+    async def test_pool_close_waits_for_release(self):
+        pool = await self.create_pool(database='postgres',
+                                      min_size=1, max_size=1)
+
+        flag = self.loop.create_future()
+        conn_released = False
+
+        async def worker():
+            nonlocal conn_released
+
+            async with pool.acquire() as connection:
+                async with connection.transaction():
+                    flag.set_result(True)
+                    await asyncio.sleep(0.1, loop=self.loop)
+
+            conn_released = True
+
+        self.loop.create_task(worker())
+
+        await flag
+        await pool.close()
+        self.assertTrue(conn_released)
+
+    async def test_pool_close_timeout(self):
+        pool = await self.create_pool(database='postgres',
+                                      min_size=1, max_size=1)
+
+        flag = self.loop.create_future()
+
+        async def worker():
+            async with pool.acquire():
+                flag.set_result(True)
+                await asyncio.sleep(0.5, loop=self.loop)
+
+        task = self.loop.create_task(worker())
+
+        with self.assertRaises(asyncio.TimeoutError):
+            await flag
+            await asyncio.wait_for(pool.close(), timeout=0.1)
+
+        await task
+
+    async def test_pool_expire_connections(self):
+        pool = await self.create_pool(database='postgres',
+                                      min_size=1, max_size=1)
+
+        con = await pool.acquire()
+        try:
+            await pool.expire_connections()
+        finally:
+            await pool.release(con)
+
+        self.assertIsNone(pool._holders[0]._con)
+
+    async def test_pool_set_connection_args(self):
+        pool = await self.create_pool(database='postgres',
+                                      min_size=1, max_size=1)
+
+        # Test that connection is expired on release.
+        con = await pool.acquire()
+        connspec = self.get_connection_spec()
+        try:
+            connspec['server_settings']['application_name'] = \
+                'set_conn_args_test'
+        except KeyError:
+            connspec['server_settings'] = {
+                'application_name': 'set_conn_args_test'
+            }
+
+        pool.set_connect_args(**connspec)
+        await pool.expire_connections()
+        await pool.release(con)
+
+        con = await pool.acquire()
+        self.assertEqual(con.get_settings().application_name,
+                         'set_conn_args_test')
+        await pool.release(con)
+
+        # Test that connection is expired before acquire.
+        connspec = self.get_connection_spec()
+        try:
+            connspec['server_settings']['application_name'] = \
+                'set_conn_args_test'
+        except KeyError:
+            connspec['server_settings'] = {
+                'application_name': 'set_conn_args_test_2'
+            }
+
+        pool.set_connect_args(**connspec)
+        await pool.expire_connections()
+
+        con = await pool.acquire()
+        self.assertEqual(con.get_settings().application_name,
+                         'set_conn_args_test_2')
+
+    async def test_pool_init_race(self):
+        pool = self.create_pool(database='postgres', min_size=1, max_size=1)
+
+        t1 = asyncio.ensure_future(pool, loop=self.loop)
+        t2 = asyncio.ensure_future(pool, loop=self.loop)
+
+        await t1
+        with self.assertRaisesRegex(
+                asyncpg.InterfaceError,
+                r'pool is being initialized in another task'):
+            await t2
+
+        await pool.close()
+
+    async def test_pool_init_and_use_race(self):
+        pool = self.create_pool(database='postgres', min_size=1, max_size=1)
+
+        pool_task = asyncio.ensure_future(pool, loop=self.loop)
+        await asyncio.sleep(0, loop=self.loop)
+
+        with self.assertRaisesRegex(
+                asyncpg.InterfaceError,
+                r'being initialized, but not yet ready'):
+
+            await pool.fetchval('SELECT 1')
+
+        await pool_task
+        await pool.close()
+
 
 @unittest.skipIf(os.environ.get('PGHOST'), 'using remote cluster for testing')
-class TestHotStandby(tb.ConnectedTestCase):
+class TestHotStandby(tb.ClusterTestCase):
     @classmethod
-    def setUpClass(cls):
-        super().setUpClass()
-
-        cls.master_cluster = cls.start_cluster(
-            pg_cluster.TempCluster,
+    def setup_cluster(cls):
+        cls.master_cluster = cls.new_cluster(pg_cluster.TempCluster)
+        cls.start_cluster(
+            cls.master_cluster,
             server_settings={
                 'max_wal_senders': 10,
                 'wal_level': 'hot_standby'
-            })
+            }
+        )
 
         con = None
 
@@ -771,27 +911,24 @@ class TestHotStandby(tb.ConnectedTestCase):
 
             conn_spec = cls.master_cluster.get_connection_spec()
 
-            cls.standby_cluster = cls.start_cluster(
+            cls.standby_cluster = cls.new_cluster(
                 pg_cluster.HotStandbyCluster,
                 cluster_kwargs={
                     'master': conn_spec,
                     'replication_user': 'replication'
-                },
+                }
+            )
+            cls.start_cluster(
+                cls.standby_cluster,
                 server_settings={
                     'hot_standby': True
-                })
+                }
+            )
 
         finally:
             if con is not None:
                 cls.loop.run_until_complete(con.close())
 
-    @classmethod
-    def tearDownMethod(cls):
-        cls.standby_cluster.stop()
-        cls.standby_cluster.destroy()
-        cls.master_cluster.stop()
-        cls.master_cluster.destroy()
-
     def create_pool(self, **kwargs):
         conn_spec = self.standby_cluster.get_connection_spec()
         conn_spec.update(kwargs)
